import os
import warnings
from typing import Callable, Dict, Union

import numpy
import numpy as np
from obspy import UTCDateTime
from obspy.io.reftek.packet import PACKET_FINAL_DTYPE

from sohstationviewer.model.mseed.read_mseed_experiment.mseed_helper import \
    Unpacker
from sohstationviewer.model.reftek.rt130_experiment.dt_packet import \
    (
    read_dt_packet, DTPacket,
)
from sohstationviewer.model.reftek.rt130_experiment.eh_et_packet import \
    (
    EHETPacket, read_eh_et_packet, eh_et_payload_end,
)
from sohstationviewer.model.reftek.rt130_experiment.reftek_helper import (
    print_error, packet_reader_placeholder, RT130ParseError, PacketHeader,
)

from sohstationviewer.model.reftek.from_rt2ms import core
from sohstationviewer.model.reftek.rt130_experiment.soh_packets import \
    (
    read_soh_packet, SOHPacket,
)


def parse_rt130_time(year: int, time_bytes: bytes) -> UTCDateTime:
    time_string = time_bytes.hex()
    # The time string has the format of DDDHHMMSSTTT, where
    # D = day of year
    # H = hour
    # M = minute
    # S = second
    # T = microsecond
    day_of_year, hour, minute, second, millisecond = (
        int(time_string[0:3]),
        int(time_string[3:5]),
        int(time_string[5:7]),
        int(time_string[7:9]),
        int(time_string[9:12])
    )
    # RT130 only stores the last two digits of the year. Because the
    # documentation for RT130 does not define a way to retrieve the full year,
    # we use Obspy's method. Accordingly, we convert 0-49 to 2000-2049 and
    # 50-99 to 1950-1999.
    if 0 <= year <= 49:
        year += 2000
    elif 50 <= year <= 99:
        year += 1900
    converted_time = UTCDateTime(year=year, julday=day_of_year, hour=hour,
                                 minute=minute, second=second,
                                 microsecond=millisecond * 1000)
    return converted_time


def get_rt130_packet_header(rt130_packet: bytes,
                            unpacker: Unpacker) -> PacketHeader:
    try:
        # Because RT130 data is always big-endian, it is more convenient to
        # use str.decode() than the unpacker.
        packet_type = rt130_packet[:2].decode('ASCII')
    except UnicodeError:
        print_error(f'Cannot decode packet type.')
        print_error('The given file does not appear to be a valid RT130 file.')
        raise RT130ParseError
    valid_packet_types = ['AD', 'CD', 'DS', 'DT', 'EH', 'ET', 'OM', 'SH', 'SC',
                          'FD']
    if packet_type not in valid_packet_types:
        print_error(f'Invalid packet type found: {packet_type}')
        print_error('The given file does not appear to be a valid RT130 file.')
        raise RT130ParseError

    experiment_number = unpacker.unpack('b', rt130_packet[2:3])[0]
    year = int(rt130_packet[3:4].hex())
    # A call to str.upper() is needed because bytes.hex() makes any
    # hexadecimal letter (i.e. ABCDEF) lowercase, while we want them to be
    # uppercase for display purpose.
    unit_id = rt130_packet[4:6].hex().upper()
    time_bytes = rt130_packet[6:12]
    packet_time = parse_rt130_time(year, time_bytes)
    byte_count = int(rt130_packet[12:14].hex())
    packet_sequence = int(rt130_packet[14:16].hex())

    return PacketHeader(packet_type, experiment_number, unit_id, packet_time,
                        byte_count, packet_sequence)


def read_rt130_file(file_name: str, unpacker: Unpacker):
    # RT130 data looks to be all big-endian (logpeek assumes this, and it has
    # been working pretty well), so we don't have to do any endianness check.

    if os.path.getsize(file_name) % 1024:
        warnings.warn('The size of the data is not a multiple of 1024; it '
                      'might be truncated.')

    packets = []

    with open(file_name, 'rb') as rt130_file:
        for i in range(os.path.getsize(file_name) // 1024):
            packet = rt130_file.read(1024)
            packet_header = get_rt130_packet_header(packet, unpacker)

            waveform_handlers: Dict[str, Callable] = {
                'EH': read_eh_et_packet,
                'ET': read_eh_et_packet,
                'DT': read_dt_packet,
            }

            soh_handlers: Dict[str, Callable] = dict.fromkeys(
                ['AD', 'CD', 'DS', 'FD', 'OM', 'SC', 'SH'], read_soh_packet
            )

            packet_handlers = {
                **waveform_handlers, **soh_handlers
            }

            packet_handler = packet_handlers.get(
                packet_header.packet_type, packet_reader_placeholder
            )
            return_val = packet_handler(packet, unpacker)
            if packet_header.packet_type == 'DT':
                packet_type = DTPacket
            elif packet_header.packet_type in ['EH', 'ET']:
                packet_type = EHETPacket
            else:
                packet_type = SOHPacket

            extended_header, data = return_val
            current_packet = packet_type(packet_header, extended_header, data)
            packets.append(current_packet)

    return packets


def convert_packet_to_obspy_format(packet: Union[EHETPacket, DTPacket],
                                   unpacker: Unpacker):
    # We want to convert the packet to a tuple. In order to make it easier to
    # maintain, we first convert the packet to a dictionary. Then, we grab the
    # values of the dictionary as tuple to get the final result.
    converted_packet = {}
    converted_packet['packet_type'] = packet.header.packet_type
    converted_packet['experiment_number'] = packet.header.experiment_number
    # Obspy only stores the last two digits of the year.
    converted_packet['year'] = packet.header.time.year % 100
    converted_packet['unit_id'] = packet.header.unit_id
    converted_packet['time'] = packet.header.time.ns
    converted_packet['byte_count'] = packet.header.byte_count
    converted_packet['packet_sequence'] = packet.header.packet_sequence
    converted_packet['event_number'] = packet.extended_header.event_number
    converted_packet[
        'data_stream_number'] = packet.extended_header.data_stream_number
    converted_packet['channel_number'] = packet.extended_header.channel_number
    converted_packet[
        'number_of_samples'] = packet.extended_header.number_of_samples
    converted_packet['flags'] = packet.extended_header.flags
    converted_packet['data_format'] = packet.extended_header.data_format

    if converted_packet['packet_type'] == 'DT':
        # Obspy stores the data as list of 1-byte integers. We store the
        # data as an arbitrary length integer, so we need to do some
        # conversion. To make converting the resulting tuple to an element
        # of a structured array of type PACKET_FINAL_DTYPE easier, we set
        # the size of the payload to be 4. This only affect data with format
        # 16, and as long as we are careful in self.to_stream, we don't even
        # have to make a special case when decoding (note: this is possible
        # because of a peculiarity of the 2's complement encoding).
        data_size = 4
        format_char = 'B'
        converted_packet['payload'] = numpy.empty(1000, np.uint8)
        packet_data = list(unpacker.unpack(
            f'{data_size}{format_char}',
            packet.data.to_bytes(data_size, 'big', signed=True)
        ))
        converted_packet['payload'][:4] = packet_data
    elif converted_packet['packet_type'] in ['EH', 'ET']:
        eh_et_payload_size = eh_et_payload_end - 24
        converted_packet['payload'] = numpy.empty(1000, np.uint8)
        packet_data = numpy.frombuffer(packet.data, np.uint8)
        converted_packet['payload'][:eh_et_payload_size] = packet_data
    else:
        converted_packet['payload'] = numpy.frombuffer(packet.data, np.uint8)
    return tuple(converted_packet.values())


class Reftek130(core.Reftek130):
    @staticmethod
    def from_file(filename: str):
        rt130_unpacker = Unpacker('>')
        rt = Reftek130()
        rt._filename = filename
        packets_in_file = read_rt130_file(filename, rt130_unpacker)
        converted_packets = []
        for packet in packets_in_file:
            converted_packets.append(
                convert_packet_to_obspy_format(packet, rt130_unpacker))
        rt._data = numpy.array(converted_packets, PACKET_FINAL_DTYPE)
        return rt


waveform_file = '/Users/kle/PycharmProjects/sohstationviewer/tests/test_data/RT130-sample/2017149.92EB/2017150/92EB/1/000000015_0036EE80'
soh_file = '/Users/kle/PycharmProjects/sohstationviewer/tests/test_data/RT130-sample/2017149.92EB/2017150/92EB/0/000000000_00000000'
import time
start = time.perf_counter()
print((Reftek130.from_file(soh_file)._data))
print('Time taken:', time.perf_counter() - start)
